import numpy as np

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.core.classes.algorithms import DataDriftDDM
from uptrain.constants import DataDriftAlgo, MeasurableType
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Monitor


class ConceptDrift(AbstractMonitor):
    dashboard_name = "concept_drift_acc"
    anomaly_type = Monitor.CONCEPT_DRIFT

    def base_init(self, fw, check):
        if check.get("measurable_args", None):
            self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
        else:
            self.measurable = MeasurableResolver(
                {"type": MeasurableType.ACCURACY}).resolve(fw)
        self.acc_arr = []
        self.avg_acc = 0
        if check["algorithm"] == DataDriftAlgo.DDM:
            warn_thres = check.get("warn_thres", 2)
            alarm_thres = check.get("alarm_thres", 3)
            self.algo = DataDriftDDM(warn_thres, alarm_thres)
        else:
            raise Exception("Data drift algo type not supported")

    def need_ground_truth(self):
        return True

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        batch_acc = self.measurable.compute_and_log(inputs, outputs, gts, extra_args)
        for acc in batch_acc:
            if acc:
                alert = self.algo.add_prediction(0)
            else:
                alert = self.algo.add_prediction(1)

            self.acc_arr.append(acc)
            self.avg_acc = (self.avg_acc * (len(self.acc_arr) - 1) + acc) / len(
                self.acc_arr
            )
            self.log_handler.add_scalars(
                "avg_accuracy",
                {"y_avg_accuracy": self.avg_acc},
                len(self.acc_arr),
                self.dashboard_name,
            )
            if isinstance(alert, str):
                self.log_handler.add_alert(
                    "Model Performance Degradation Alert 🚨",
                    alert,
                    self.dashboard_name
                )


import numpy as np
from uptrain.core.lib.helper_funcs import cluster_and_plot_data

class Clustering:
    def __init__(self, args) -> None:
        self.NUM_BUCKETS = args["num_buckets"]
        self.is_embedding = args["is_embedding"]
        self.plot_save_name = args.get("plot_save_name", "")
        self.cluster_plot_func = args.get('cluster_plot_func', None)
        self.dist = []
        self.dist_counts = []
        self.max_along_axis = []
        self.low_density_regions = []
        self.idxs_closest_to_cluster_centroids = {}


    def cluster_data(self, data):
        if self.is_embedding:
            self.bucket_vector(data)
        else:
            buckets = []
            clusters = []
            cluster_vars = []
            for idx in range(data.shape[1]):
                this_inputs = data[:, idx]
                this_buckets, this_clusters, this_cluster_vars = self.bucket_scalar(
                    this_inputs
                )
                buckets.append(this_buckets)
                clusters.append(this_clusters)
                cluster_vars.append(this_cluster_vars)
            self.buckets = np.array(buckets)
            self.clusters = np.array(clusters)
            self.cluster_vars = np.array(cluster_vars)

        self.dist = np.array(self.dist)
        self.dist_counts = np.array(self.dist_counts)

        clustering_results = {
            "buckets": self.buckets,
            "clusters": self.clusters,
            "cluster_vars": self.cluster_vars,
            "dist": self.dist,
            "dist_counts": self.dist_counts,
            "max_along_axis": self.max_along_axis,
            'low_density_regions': self.low_density_regions,
            "idxs_closest_to_cluster_centroids": self.idxs_closest_to_cluster_centroids
        }

        return clustering_results


    def bucket_scalar(self, arr):
        if isinstance(arr[0], str):
            uniques, counts = np.unique(np.array(arr), return_counts=True)
            buckets = uniques
            self.NUM_BUCKETS = len(buckets)
            clusters = uniques
            cluster_vars = [None] * self.NUM_BUCKETS
            self.ref_dist.append([[counts[x] / len(arr)] for x in range(self.NUM_BUCKETS)])
            self.ref_dist_counts.append(
                [[counts[x]] for x in range(self.NUM_BUCKETS)]
            )
        else:
            sorted_arr = np.sort(arr)
            buckets = []
            clusters = []
            cluster_vars = []
            for idx in range(0, self.NUM_BUCKETS):
                if idx > 0:
                    buckets.append(
                        sorted_arr[int(idx * (len(sorted_arr) - 1) / self.NUM_BUCKETS)]
                    )
                this_bucket_elems = sorted_arr[
                    int((idx) * (len(sorted_arr) - 1) / self.NUM_BUCKETS) : int(
                        (idx + 1) * (len(sorted_arr) - 1) / self.NUM_BUCKETS
                    )
                ]
                gaussian_mean = np.mean(this_bucket_elems)
                gaussian_var = np.var(this_bucket_elems)
                clusters.append([gaussian_mean])
                cluster_vars.append([gaussian_var])

        self.dist.append([[1 / self.NUM_BUCKETS] for x in range(self.NUM_BUCKETS)])
        self.dist_counts.append(
            [[int(len(sorted_arr) / self.NUM_BUCKETS)] for x in range(self.NUM_BUCKETS)]
        )
        return np.array(buckets), np.array(clusters), np.array(cluster_vars)

    def bucket_vector(self, data):

        abs_data = np.abs(data)
        self.max_along_axis = np.max(abs_data, axis=0)
        data = data/self.max_along_axis

        all_clusters, counts, cluster_vars, density_around_points, idxs_closest_to_cluster_centroids = cluster_and_plot_data(
            data,
            self.NUM_BUCKETS,
            cluster_plot_func=self.cluster_plot_func,
            plot_save_name=self.plot_save_name,
            normalisation=self.max_along_axis
        )
        low_density_regions = data[np.where(density_around_points < np.ceil(len(density_around_points) * 0.002))[0]]

        self.clusters = np.array([all_clusters])
        self.cluster_vars = np.array([cluster_vars])
        self.buckets = self.clusters
        self.low_density_regions = low_density_regions
        self.density_around_points = density_around_points
        self.idxs_closest_to_cluster_centroids = idxs_closest_to_cluster_centroids

        self.dist_counts = np.array([counts])
        self.dist = self.dist_counts / data.shape[0]

    def infer_cluster_assignment(self, feats, prod_dist_counts=None):
        if prod_dist_counts is None:
            prod_dist_counts = np.zeros((feats.shape[1], self.NUM_BUCKETS))
        if self.is_embedding:
            selected_cluster = np.argmin(
                np.sum(
                    np.abs(self.clusters[0] - feats),
                    axis=tuple(range(2, len(feats.shape))),
                ),
                axis=1,
            )
            for clus in selected_cluster:
                prod_dist_counts[0][clus] += 1
            this_datapoint_cluster = selected_cluster
        else:
            this_datapoint_cluster = []
            for idx in range(feats.shape[2]):
                if isinstance(feats[0,0,idx], str):
                    try:
                        bucket_idx = np.array([
                            list(self.buckets[idx]).index(feats[x,0,idx]) for x in range(feats.shape[0])
                        ])
                    except:
                        # TODO: This logic is not completely tested yet. Contact us if you are facing issues
                        # If given data-point is not present -> add a new bucket
                        temp_buckets = list(self.buckets[idx])
                        num_added = 0

                        for x in range(feats.shape[0]):
                            if feats[x,0,idx] not in temp_buckets:
                                temp_buckets.append(feats[x,0,idx])
                                num_added += 1
                        
                        self.buckets[idx] = np.array(temp_buckets)
                        bucket_idx = np.array([
                            list(self.buckets[idx]).index(feats[x,0,idx]) for x in range(feats.shape[0])
                        ])
                else:
                    bucket_idx = np.searchsorted(
                        self.buckets[idx], feats[:, :, idx]
                    )[:, 0]
                this_datapoint_cluster.append(bucket_idx)
                for clus in bucket_idx:
                    prod_dist_counts[idx][clus] += 1
            this_datapoint_cluster = np.array(this_datapoint_cluster)
        return this_datapoint_cluster, prod_dist_counts


import numpy as np


class DataDriftDDM:
    """
    Using the DDM data drift detection method
    to detect drift on dataset coming from a
    Binomial distribution.
    Args:
    warn_thres: Threshold for drift warning
    alarm_thres: Threshold for drift detection alarm
    """

    def __init__(self, warn_thres, alarm_thres):
        super().__init__()
        self.warn_thres = warn_thres
        self.alarm_thres = alarm_thres

        # Total number of predictions by the distribution
        self.total_count = 1
        # If drift is detected
        self.drift_detected = False

        self._reset()

    def _reset(self):
        """Reset the statistics."""
        # Number of predictions before the next drift
        self.current_count = 1
        # Estimated mean of the Binomial distribution
        self.mean = 1
        # Estimated standard deviation of the Binomial distribution
        self.std = 0
        # If drift is in warning zone
        self.is_warning_zone = False
        self.ms_min = np.Inf
        self.mean_min = np.Inf
        self.std_min = np.Inf
        self.alert = None

    def add_prediction(self, prediction):
        """
        Checks for data drift in the binomial distribution
        Parameters
        ----------
        prediction : Current prediction by the Binomial distribution

        Returns
        -------
        drift_detected : boolean
        """
        if self.drift_detected is True:
            # self._reset()
            return

        self.mean += (prediction - self.mean) / self.current_count
        self.std = np.sqrt(self.mean * (1 - self.mean) / self.current_count)

        self.current_count += 1
        self.total_count += 1
        self.drift_detected = False

        if self.current_count < 500:
            return False

        if self.mean + self.std <= self.ms_min:
            self.mean_min = self.mean
            self.std_min = self.std
            self.ms_min = self.mean + self.std

        if self.mean + self.std > self.mean_min + self.alarm_thres * self.std_min:
            self.drift_detected = True
            self.alert = f"Drift detected with DDM at time: {self.total_count}!!!"
            print(self.alert)
        elif self.mean + self.std > self.mean_min + self.warn_thres * self.std_min:
            self.is_warning_zone = True
        else:
            self.is_warning_zone = False
        return self.alert
